----------------------------------------------------------------------------------
-- Company: University of Ljubljana, Faculty of Electrical Engineering
-- Engineer: Andrej Trost
-- 
-- Create Date: 03/10/2015 01:15:47 PM
-- Design Name: Red Pitaya Graphics
-- Module Name: red_pitaya_graph
-- Project Name: http://www.redpitaya.com
-- Description: 
-- 
-- Revision: 1
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

entity red_pitaya_graf is
    Port ( clk50 : in STD_LOGIC; -- VGA signals
           hsync : out STD_LOGIC;
           vsync : out STD_LOGIC;
           rgb   : out STD_LOGIC_VECTOR (2 downto 0);
           
           -- ext_io serial interface
           ld  : out STD_LOGIC;
           sin : in STD_LOGIC;
           
           led : out STD_LOGIC_VECTOR (3 downto 0);
           -- acquisition signals (red pitaya scope buffer)
           ack_rp   : out STD_LOGIC_VECTOR (13 downto 0); -- read pointer
           ack_scale: out STD_LOGIC_VECTOR (1 downto 0); -- scale graph
           ack_data : in STD_LOGIC_VECTOR (13 downto 0);  -- data bus
           ack_datb : in STD_LOGIC_VECTOR (13 downto 0);  -- data bus

           -- System bus
           clk_i:   in STD_LOGIC;                      -- bus clock 
           rstn_i:  in STD_LOGIC;                      -- bus reset - active low
           adr_i:   in STD_LOGIC_VECTOR(31 downto 0);  -- bus address
           wdata_i: in STD_LOGIC_VECTOR(31 downto 0);  -- bus write data          
           wen_i:   in STD_LOGIC;                      -- bus write enable
           rdata_o: out STD_LOGIC_VECTOR(31 downto 0); -- bus read data
           err_o:   out STD_LOGIC;                     -- bus error indicator
           ack_o:   out STD_LOGIC                      -- bus acknowledge signal         
           );
end red_pitaya_graf;

architecture Behavioral of red_pitaya_graf is

 -- Graphics sinchronization and ext timing control
 component VGA is
     Port ( clk50 : in STD_LOGIC;
        hsync : out STD_LOGIC;
        vsync : out STD_LOGIC;
        rgb : out STD_LOGIC_VECTOR (2 downto 0);
        cx : out unsigned (9 downto 0);
        cy : out unsigned (9 downto 0);
        en : out STD_LOGIC;
        oe : out STD_LOGIC;
        sck : out STD_LOGIC;
        ldcp : out STD_LOGIC);
 end component;

 -- External interface serialize/deserialize, decode rotary encoder
 -- and generate events: increment, decrement, key press
 component ext_interface is
    Port ( clk : in STD_LOGIC;
        sck : in STD_LOGIC;
        sin : in STD_LOGIC;
        ldcp : in STD_LOGIC;           
		inc : out STD_LOGIC;
		dec : out STD_LOGIC;
        key : out STD_LOGIC_VECTOR (1 downto 0)        
        ); 
 end component;

 -- Graphics control from external events: scale, vertical gain and position
 component ext_control is
    Port ( clk : in STD_LOGIC;
		inc : in STD_LOGIC;
		dec : in STD_LOGIC;
        key : in STD_LOGIC_VECTOR (1 downto 0);
		mode: out STD_LOGIC_VECTOR (4 downto 0);
		scale: out unsigned (2 downto 0); -- "000"-"101", 512, 1k, 2k, 4k, 8k, 16k
		vert1, vert2: out unsigned (2 downto 0); -- vertical scale "000"-"101"
		pos1, pos2:  out unsigned (7 downto 0) -- use pos1 & "00"
        ); 
 end component;

 -- Character ROM
 component CROM is
    Port ( adr : in UNSIGNED (8 downto 0);
           data : out UNSIGNED (7 downto 0));
 end component;

 type logo is array(0 to 19) of std_logic_vector(0 to 79);
 signal rplogo: logo := (
 "00000000100000000000000000000000000000000000000000000000000000000000000000000000",
 "00000000100000000000000000000000000000000000000000000000000000000000000000000000",
 "00000001100000000000000000000000000000000000000000000000000000000000000000000000",
 "00000001000000000000000000000000000000000000000000000000000000000000000000000000",
 "00000001001000000000000000000000000000000000000000000000000000000000000000000000",
 "00000001001000000000000000000000000000100000000000000000000000000000000000000000",
 "00000001111000000000000000000000000000100000000001001000000000000000000000000000",
 "00001001111000000000000000000000000000100000000000001000000000000000000000000000",
 "00010011111001000001011001111100011111100011110001011111011111100100000101111110",
 "00011111111111000001100010000010100000100100001001001000000000010100000100000001",
 "00011111111110000001000010000010100000100100001001001000000000010100000100000001",
 "00011111111111000001000010000010100000100100001001001000001111110100000100111111",
 "01111110000000000001000011111000100000100100001001001000010000010100000101000001",
 "00111110011001100001000010000000100000100100001001001000010000010100000101000001",
 "00111100011001100001000010000000100000100100001001001000010000010100000101000001",
 "00011110000000000001000010000000100000100100001001001000010000010100000101000001",
 "00011111111111000001000001111100011111100111110001000111001111110011111100111111",
 "00011111111110000000000000000000000000000100000000000000000000000000000100000000",
 "00001111111110000000000000000000000000000100000000000000000000000000000100000000",
 "00000011111000000000000000000000000000000100000000000000000000000111111000000000");
 
 signal logox : unsigned(7 downto 0);
 
 signal cx,cy : unsigned (9 downto 0);
 signal en, oe: std_logic;
 signal rgb0 : std_logic_vector(2 downto 0);
 
 signal adr: unsigned(8 downto 0); 
 signal znak: unsigned(7 downto 0);
  
 type ram1k is array(0 to 255) of std_logic_vector(31 downto 0);
 signal txtram: ram1k;
 signal txtadr : unsigned(9 downto 0);
 signal txtdata: std_logic_vector(31 downto 0);
 signal ascii: unsigned(6 downto 0);
 
 type ramgraf is array(0 to 511) of signed(7 downto 0);
 signal ramg: ramgraf;
 signal g0adr, g0_x : unsigned(8 downto 0);
 signal g0_y, g1_y, g2_y : signed(7 downto 0);
 signal g0_yc : integer;
 signal g0_pos: STD_LOGIC_VECTOR (7 downto 0);
  
 constant sat1: signed(7 downto 0) := "01111111";
 constant sat0: signed(7 downto 0) := "10000000";
 
 signal g1abs,g1pix,g1pix0,g1min,g1max: unsigned(8 downto 0);
 signal g2abs,g2pix,g2pix0,g2min,g2max: unsigned(8 downto 0);
 
 signal sck,ldcp: std_logic;
 signal g1_pos: STD_LOGIC_VECTOR (7 downto 0); 
 signal g1_scale: STD_LOGIC_VECTOR (2 downto 0);
 
 signal inc, dec: std_logic;
 signal key : STD_LOGIC_VECTOR (1 downto 0);
 signal mode: STD_LOGIC_VECTOR (4 downto 0);
 signal scale: unsigned(2 downto 0);
 signal pos1, pos2: unsigned (7 downto 0);
 signal vert1, vert2: unsigned (2 downto 0);
 
 signal tmp_rp: STD_LOGIC_VECTOR (13 downto 0);
begin

u0: VGA port map (
 clk50 => clk50,
 hsync => hsync,
 vsync => vsync,
 rgb => rgb0,
 cx => cx,
 cy => cy,
 en => en,
 oe => oe,
 sck => sck,
 ldcp => ldcp );


-- READ characters
 txtadr <= cy(7 downto 3) & cx(7 downto 3);

 txtdata <= txtram(to_integer(txtadr(9 downto 2)));


 with txtadr(1 downto 0) select
 ascii <= unsigned(txtdata(6 downto 0)) when "00",
          unsigned(txtdata(14 downto 8)) when "01",
          unsigned(txtdata(22 downto 16)) when "10",
          unsigned(txtdata(30 downto 24)) when others;       
 
 adr <= resize(ascii-32,6) & cy(2 downto 0);

u1: CROM port map (adr => adr, data => znak); 

--sck <=  '1' when cx>=856 and cx<976 else '0'; 
ld <= ldcp; -- external

u2: ext_interface port map (
 clk => clk50,
 sck => sck,
 sin => sin,
 ldcp => ldcp,
 inc => inc,
 dec => dec,
 key => key
);

u3: ext_control port map (
 clk => clk50,
 inc => inc,
 dec => dec,
 key => key,
 mode => mode,
 scale => scale,
 vert1 => vert1,
 vert2 => vert2,
 pos1 => pos1,
 pos2 => pos2
);

led <= mode(4 downto 1); --pos(3 downto 0);

-- graph logic
 ack_scale <= "00"; --std_logic_vector(scale(1 downto 0));

 g0_y <= ramg(to_integer(cx+g0_x-256)); -- relativna y koordinata
 g0_yc <= (128 + to_integer(unsigned(g0_pos))) - to_integer(cy);

 -- horizontal scale
 tmp_rp <= std_logic_vector(resize(cx-256, 14));
 with scale select
   ack_rp <= tmp_rp(12 downto 0) & '0' when "001",
             tmp_rp(11 downto 0) & "00" when "010",
             tmp_rp(10 downto 0) & "000" when "011",
             tmp_rp(9 downto 0) & "0000" when "100",
             tmp_rp(8 downto 0) & "00000" when "101",
             tmp_rp when others;   
             
 -- vertical gain (shift data, add saturation!)
 psat: process(ack_data, ack_datb, vert1, vert2)
 begin
   case vert1 is  
      when "010" =>
        if ack_data(13 downto 12)="01" then g1_y <= sat1;
        elsif ack_data(13 downto 12)="10" then g1_y <= sat0;
        else g1_y <= signed(ack_data(12 downto 5)); end if;
      when "011" =>
        if ack_data(13)='0' then -- positve
          if ack_data(12 downto 11)="00" then g1_y <= signed(ack_data(11 downto 4));
          else g1_y <= sat1; end if;
        else
          if ack_data(12 downto 11)="11" then g1_y <= signed(ack_data(11 downto 4));
          else g1_y <= sat0; end if;
        end if; 
      when "100" =>
          if ack_data(13)='0' then -- positve
            if ack_data(12 downto 10)="000" then g1_y <= signed(ack_data(10 downto 3));
            else g1_y <= sat1; end if;
          else
            if ack_data(12 downto 10)="111" then g1_y <= signed(ack_data(10 downto 3));
            else g1_y <= sat0; end if;
          end if; 
      when "101" =>
          if ack_data(13)='0' then -- positve
            if ack_data(12 downto 9)="0000" then g1_y <= signed(ack_data(9 downto 2));
            else g1_y <= sat1; end if;
          else
            if ack_data(12 downto 9)="1111" then g1_y <= signed(ack_data(9 downto 2));
            else g1_y <= sat0; end if;
          end if;        
      when others => g1_y <= signed(ack_data(13 downto 6)); -- "000" or "101"       
   end case;
   
   case vert2 is  
       when "010" =>
         if ack_datb(13 downto 12)="01" then g2_y <= sat1;
         elsif ack_datb(13 downto 12)="10" then g2_y <= sat0;
         else g2_y <= signed(ack_datb(12 downto 5)); end if;
       when "011" =>
         if ack_datb(13)='0' then -- positve
           if ack_datb(12 downto 11)="00" then g2_y <= signed(ack_datb(11 downto 4));
           else g2_y <= sat1; end if;
         else
           if ack_datb(12 downto 11)="11" then g2_y <= signed(ack_datb(11 downto 4));
           else g2_y <= sat0; end if;
         end if; 
       when "100" =>
           if ack_datb(13)='0' then -- positve
             if ack_datb(12 downto 10)="000" then g2_y <= signed(ack_datb(10 downto 3));
             else g2_y <= sat1; end if;
           else
             if ack_datb(12 downto 10)="111" then g2_y <= signed(ack_datb(10 downto 3));
             else g2_y <= sat0; end if;
           end if; 
       when "101" =>
           if ack_datb(13)='0' then -- positve
             if ack_datb(12 downto 9)="0000" then g2_y <= signed(ack_datb(9 downto 2));
             else g2_y <= sat1; end if;
           else
             if ack_datb(12 downto 9)="1111" then g2_y <= signed(ack_datb(9 downto 2));
             else g2_y <= sat0; end if;
           end if;        
       when others => g2_y <= signed(ack_datb(13 downto 6)); -- "000" or "101"       
   end case;
 end process;

 -- calculate abscissa and pixel position: pixel = abscissa - data
 g1abs <= to_unsigned(128+32, 9) + pos1; 
 g1pix <= g1abs - unsigned(resize(signed(g1_y), 9));

 g2abs <= to_unsigned(128+32, 9) + pos2; 
 g2pix <= g2abs - unsigned(resize(signed(g2_y), 9));
 
 pdel: process(clk50)
 begin
  if rising_edge(clk50) then   
    g1pix0 <= g1pix;
    g2pix0 <= g2pix;  -- save last pixel    
  end if;
 end process;
 
 -- draw line from min to max
 g1min <= g1pix when g1pix<g1pix0 else g1pix0;
 g1max <= g1pix when g1pix>g1pix0 else g1pix0;

 g2min <= g2pix when g2pix<g2pix0 else g2pix0;
 g2max <= g2pix when g2pix>g2pix0 else g2pix0;
 
p: process(clk50)
begin
 if rising_edge(clk50) then   
   if en='1' then
     rgb <= "000"; -- black background
     -- logo(0,0)-(80,20)
     if cx >= logox and cx < 80+logox and cy < 20 then 
       if rplogo(to_integer(cy))(to_integer(cx-logox))='1' then
         if (cx-logox)>40 then rgb <= "111"; else rgb <= "100"; end if;
       end if;
     -- text field, yellow text
     elsif cx(9 downto 3) >= 0 and cx(9 downto 3) < 32 and cy(9 downto 3)>=16 and cy(9 downto 3)<48 then
       if znak(to_integer(7 - cx(2 downto 0)))= '1' then
         rgb <= "110";   
       else
         rgb <= "000";
       end if;
     elsif cy = 20 then -- horizontal line
        rgb <= "111";
     -- Graph area
     elsif cx >= 256 and cx < 768 and cy >= 32 and cy < 512+32 then
	   rgb <= "000";
	   -- graph axes
	   if cx=512 or cy=288 then 
	     rgb <= "111";
	   elsif oe='1' and (cx=262 or cx=387 or cx=637 or cx=762 or
	         cy=51  or cy=169 or cy=407 or cy=525) then
			 rgb <= "111";
	   end if;
	   if (vert1/="101") and cx(2)='0' and cy=g1abs then
	     rgb <= "100";
	   end if;
	   if (vert2/="101") and cx(2)='0' and cy=g2abs then
         rgb <= "110";
       end if;
	   
       if (g0_y  = g0_yc) then 
         rgb <= "001"; -- blue SW graph
       elsif (vert1/="101") and (cy>=g1min and cy<=g1max) then
         rgb <= "100"; -- red channel 1
       elsif (vert2/="101") and (cy>=g2min and cy<=g2max) then
         rgb <= "110"; -- yellow channel 2
       end if;
--       if cy=32+128 then --cy=32+128
--         rgb <= "001";
--       end if;              
     else     -- mavrica spodaj levo
      if cx < 256 and cy > 512 and cy < 544 then
        rgb <= std_logic_vector(cx(7 downto 5));      
      end if;
     end if;
   else
     rgb <= "000";
   end if;
 end if;
end process;



-- bus interface logic
pbus: process(clk_i)
begin
 if rising_edge(clk_i) then
   if wen_i='1' then -- dodal wen_i
     if adr_i(19 downto 0) = X"00008" then
       logox <= unsigned(wdata_i(7 downto 0));
     elsif adr_i(19 downto 0) = X"0000C" then
       g0_pos <= wdata_i(7 downto 0);     
     elsif adr_i(19 downto 0) = X"00020" then
       g0adr <= g0adr + 1;
       ramg(to_integer(g0adr)) <= signed(wdata_i(7 downto 0));
       g0_x <= g0_x + 1;
     elsif adr_i(19 downto 10) = "1000000000" then
       txtram( to_integer(unsigned( adr_i(9 downto 2)))) <= wdata_i; -- pii v txt ram 
     end if;
   end if;
 end if;
end process;

err_o <= '0';

pbusr: process(adr_i, mode, scale, vert1, vert2)
begin   
  case adr_i(19 downto 0) is
    when X"00000" => rdata_o <= X"000000" & "000" & mode; ack_o <= '1';
    when X"00010" => rdata_o <= X"0000000" & '0' & std_logic_vector(scale); ack_o <= '1';
    when X"00014" => rdata_o <= X"0000000" & '0' & std_logic_vector(vert1); ack_o <= '1';
    when X"00018" => rdata_o <= X"0000000" & '0' & std_logic_vector(vert2); ack_o <= '1';    
    when others => rdata_o <= (others=>'0'); ack_o <= '1';  
  end case;
--  rdata_o <= (others=>'0'); ack_o <= '1';
end process;

end Behavioral;
